
TaskProxyEditor {
	classvar <>scheme; 

	var <>proxy, <>nVars, <win, <zone, <flow, <skipjack, <usesPlayN; 
	var <nameBut, <playBut, <pauseBut, <srcBut, <envBut, <varButs;
	var <usedKeys, <oldState = #[];
	
	*new { |px, nVars=0, height=16, width = 30, w, makeWatcher=true| 
		^super.new.proxy_(px).nVars_(nVars)
			.init(w, height, width, makeWatcher); 
	}
	
	*initClass { 	// scheme defines the class look and feel.
		scheme = ();
		scheme.font = GUI.scheme.font.new("Monaco", 10);
		scheme.fontColor = Color.black;
		
		scheme.background = Color(0.8, 0.85, 0.7, 0.5);
		scheme.foreground = Color.grey(0.95);
		scheme.onColor = Color(0.5, 1, 0.5); 
		scheme.offColor = Color.clear; 
		
		scheme.gap = 0 @ 0;
	}

	srcString {
		^this.observedClass.asString 
			++ "(" + proxy.key.asCompileString + "," 
			+ proxy.source.asCompileString + ");\n"
	}

	editString { |edKey| 
		var keyText = if (edKey.isNil) { 
			"\anyKey, nil"
		} { 
			edKey.asCompileString + "," 
			+ proxy.envir[edKey].asCompileString 
		};
		
		^(this.observedClass.asString 
			++ "(" + proxy.key.asCompileString 
			+ ").set(" + keyText + ");\n"
		)
	}

	editStrings { |edKeys| 
		edKeys = edKeys ? this.getUsedKeys;
		^edKeys.collect (this.editString(_))
	}

	getUsedKeys { 
		if (proxy.envir.isNil) { ^[] };
		^usedKeys = proxy.envir.keys.rejectAs(_ == \self, Array);
	}

	openDoc { |strings, bounds|
		Document("edit me", strings.join ).bounds_(bounds ? Rect(0, 400, 400, 200));
	}

	init { |w, height, width, makeWatcher| 
		var sys = GUI.scheme; 
		scheme.font = sys.font.new(scheme.font.name, scheme.font.size);

		win = w ?? { sys.window.new("task px edit", Rect(0, 0, 190, 30)) };
		zone = sys.compositeView.new(win, Rect(0, 0, 190, height)); 
		zone.background_(scheme.foreground);
		
		flow = FlowLayout(zone.bounds, 0@0, scheme.gap);
		zone.decorator = flow; 
		try { win.front };

		nameBut = sys.button.new(zone, Rect(0,0, 60,height))
			.font_(scheme.font)
			.states_([[" ", scheme.fontColor, scheme.offColor]])
			.action_({ "edit all".postln })
			.keyDownAction_({ |btn, char| if (char.ascii == 127) { proxy.clear }; })
		;

		playBut = sys.button.new(zone, Rect(0,0, width, height))
			.font_(scheme.font)
			.states_([
				[" >", scheme.fontColor, scheme.offColor], 
				[" _", scheme.fontColor, scheme.onColor ]
			])
			.action_({ |b| 
				[ { proxy.stop }, { proxy.play } ][b.value].value 
			});

		pauseBut = sys.button.new(zone, Rect(0,0, width, height))
			.font_(scheme.font)
			.states_([
				["paus", scheme.fontColor, scheme.onColor], 
				["rsum", scheme.fontColor, scheme.offColor]
			])
			.action_({ |b| 
				[ { proxy.resume }, { proxy.pause } ][b.value].value 
			});

		srcBut = sys.button.new(zone, Rect(0,0, width, height))
			.font_(scheme.font)
			.states_([ ["src", scheme.fontColor, scheme.offColor] ])
			.action_({ 
				this.openDoc(this.srcString)
			});

		envBut = sys.button.new(zone, Rect(0,0, width, height))
			.font_(scheme.font)
			.states_([ ["env", scheme.fontColor, scheme.offColor] ])
			.action_({ 
				if (proxy.envir.isNil) { 
					this.openDoc(this.editString)
				} { 
					this.openDoc(this.editStrings)
				};
			});
//					// one key per button
//		varButs = Array.fill(nVars, { |i| 
//			sys.button.new(zone, Rect(0,0, width, height))
//				.font_(scheme.font)
//				.states_([ ["-", scheme.fontColor, scheme.offColor] ])
//				.action_({ |b| 
//					this.openDoc(this.editString(b.states[0][0].asSymbol))
//				});
//		});
		if (makeWatcher) { this.makeWatcher };
	}

	makeWatcher { 
		skipjack.stop;
		skipjack = SkipJack({ this.updateAll }, 
			0.5, 
			{ win.isClosed },
			this.observedClass.name.asString + try { proxy.key }
		);
		skipjack.start;
	}
	updateAll { 
		var proxyChanged, newState;

		if (win.isClosed) {skipjack.stop; ^this };

		if (proxy.isNil) { 
			[ nameBut, playBut, pauseBut, srcBut, envBut ].flat.do (_.visible_(false));
//			nameBut.states.first.put(0, "");
//			playBut.value_(0);
//			pauseBut.value_(0);
//			varButs.do { |btn| btn.states.first.put(0, "-") };
			win.refresh;
			oldState = [];
			^this;
		}; 

		// get newState of Tdef: 
		newState = [ 
			proxy, 
			proxy.notNil and: { proxy.key }, proxy.isPlaying, 
			proxy.player.notNil and: { proxy.player.wasStopped } ];
		
		// compare with oldState, if differences, do partial update. 
		if (oldState == newState) { /*"nothing doing".postln;*/ ^this }; 
		
		if (newState != oldState) {
			[ nameBut, playBut, pauseBut, srcBut, envBut ].flat.do (_.visible_(true));
			"changed!".postln; 
			nameBut.states_([[newState[1]]]);
			playBut.value_(newState[2].binaryValue);
			pauseBut.value_(newState[3].binaryValue);
			win.refresh;
		};
		
		// every 100th or so, do fullUpdate anyway. 
				
		oldState = newState;	// keep for next check
	}
	clear { proxy = nil }
	

}

TdefEditor : TaskProxyEditor { 
	observedClass { ^Tdef }
}

PdefEditor : TaskProxyEditor { 
	observedClass { ^Pdef }
}

TdefAllGui { 
	var <win, <edits, <names; 
	*new { |size=32, w| 
		^super.new.init(size, w);
	}
	init { | size, w|
		var zone, flow;
		var sys = GUI.scheme; 
		win = w ?? { sys.window.new("Tdef-all", Rect(0, 0, 194, 16 * size + 5)) };
		zone = sys.compositeView.new(win, Rect(0, 0, 200, 16 * size + 10)); 
	//	zone.background_(scheme.foreground);

		flow = FlowLayout(zone.bounds, 0@0, 0@0);
		zone.decorator = flow; 
		try { win.front };
		
		edits = Array.fill(size, { TdefEditor(w: zone) }); 
		
		SkipJack({ 
			names = Tdef.all.keys.as(Array).sort.keep(size); 
			edits.do { |edit, i| edit.proxy_(Tdef.all[names[i]]) };
		}, 1, { win.isClosed }, "tdefall");
	}
}